-------------------------------------------------------------------------------
--
-- Video Decoder Peripheral
--
--
--
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
library UNISIM;
use UNISIM.VComponents.all;


entity opb_video_decoder is

  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"0180_0000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"0180_3FFF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(31 downto 0);
    OPB_BE      : in  std_logic_vector(3 downto 0);
    OPB_DBus    : in  std_logic_vector(31 downto 0);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    VID_DBus    : out std_logic_vector(31 downto 0);
    VID_errAck  : out std_logic;        -- (unused)
    VID_retry   : out std_logic;        -- (unused)
    VID_toutSup : out std_logic;        -- Timeout suppress
    VID_xferAck : out std_logic;       -- Transfer acknowledge

   -- Coming from SAA7114H
    IPD         : in std_logic_vector (7 downto 0);
    HPD         : in std_logic_vector (7 downto 0);
    IDQ         : in std_logic;
    ICLK        : in std_logic;
    IPGV        : in std_logic;
    IPGH        : in std_logic;
    ITRI        : out std_logic;
    ITRDY       : out std_logic
  );
end opb_video_decoder;

architecture structural of opb_video_decoder is

subtype counts is std_logic_vector(15 downto 0);
type reg_num is array(15 downto 0) of counts;

signal grid_num : reg_num;


-- Buffered version of the signals
-- with the same name in the entity
signal buf_iclk : std_logic;
signal buf_ipgh : std_logic;
signal buf_ipgv : std_logic;
signal buf_idq : std_logic;
signal buf_ipd : std_logic_vector (7 downto 0);
signal buf_hpd : std_logic_vector (7 downto 0);
signal buf_itri : std_logic;
signal buf_itrdy : std_logic;

-- Latched versions of the above buffered signals
signal latched_ipgh  : std_logic;
signal latched_ipgv  : std_logic;
signal latched_idq   : std_logic;
signal latched_ipd : std_logic_vector (7 downto 0);
signal latched_hpd : std_logic_vector (7 downto 0);

-- Signals used when reading to interface
-- with OPB bus
signal cs : std_logic;
signal ce : std_logic;
signal rnw : std_logic;
signal xfer : std_logic;

-- signal to read values from hardware
signal raddr : std_logic_vector(11 downto 0);

signal filling_level : std_logic_vector(3 downto 0);

-- Count the number of lines read from video decoder
signal line_count : std_logic_vector(15 downto 0);

-- Count the frame (Actually, it's the frame ID
signal frame_counter : std_logic_vector(1 downto 0);

-- Data coming from video decoder interface :
signal data_from_decoder : std_logic_vector(15 downto 0);

-- Data bus and latched data bus
signal data_from_regs : std_logic_vector (15 downto 0);
signal data_bus_ce : std_logic_vector (31 downto 0);

signal load_regs : std_logic_vector (16 downto 0);
signal load_data : std_logic_vector (3 downto 0);
signal load : std_logic_vector (4 downto 0);

-- Signals for the memory read state machine
signal q2, q1, q0 : std_logic;

-- Coming from video_processor, going to registers
signal intf_idq_out  : std_logic;
signal intf_iclk_out : std_logic;
signal ICLK_OUT        : std_logic;
signal block_count     : std_logic_vector (15 downto 0);

signal waddr : std_logic_vector(10 downto 0);
signal pix_count          : std_logic_vector(10 downto 0);
signal active             : std_logic;
signal threshold          : std_logic_vector(31 downto 0);


component video_decoder_intf is
 port (
    ipd          : in std_logic_vector (7 downto 0);
    hpd          : in std_logic_vector (7 downto 0);
    idq          : in std_logic;
    iclk_in      : in std_logic;
    ipgh         : in std_logic;
    ipgv         : in std_logic;
    data         : out std_logic_vector (15 downto 0);
    waddr        : out std_logic_vector (10 downto 0);
    idq_out      : out std_logic;
    iclk_out     : out std_logic;
    fil_level    : out std_logic_vector(3 downto 0);
    line_count   : out std_logic_vector(15 downto 0);
    frame_id     : out std_logic_vector(1 downto 0);
    pixel_counter: out std_logic_vector (10 downto 0);
    active_out   : out std_logic;
    reset        : in std_logic
  );
end component;

component  video_processing  is
  port (
    IPD              : in  std_logic_vector(7 downto 0);
    HPD              : in  std_logic_vector(7 downto 0);
    IDQ              : in  std_logic;
    ICLK             : in  std_logic;
    ICLK_OUT         : out std_logic;
    pix_count        : in  std_logic_vector(10 downto 0);
    line_count       : in  std_logic_vector(15 downto 0);
    block_count      : out std_logic_vector(15 downto 0);
    load             : out  std_logic_vector(4 downto 0);
    reset            : in  std_logic;
    threshold        : in  std_logic_vector(31 downto 0)
    );
end component ;

component demux_1_16 is
    port( sels : in std_logic_vector(4 downto 0);
          output: out std_logic_vector (16 downto 0));
end component;

component mux_16_1 is
port(
  i0, i1, i2, i3, i4, i5, i6, i7 : in std_logic_vector(15 downto 0);
  i8, i9, i10, i11, i12, i13, i14, i15: in std_logic_vector(15 downto 0);
  sels : in std_logic_vector (3 downto 0);
  z : out std_logic_vector(15 downto 0));
end component;

component regis16 is
    port( rst, clk, load: in std_logic;
	  input: in std_logic_vector( 15  downto 0 );
	  output: out std_logic_vector( 15 downto 0 )
    );
end component;

component IBUFG is
  port (
    I : in  std_logic;
    O : out std_logic);
end component;

component IBUF
  port (
    I : in  STD_ULOGIC;
    O : out STD_ULOGIC);
end component;

component OBUF
  port(
    O:	out std_ulogic;
    I:	in  std_ulogic
    );
end component;

component FD
  port (
    C : in std_logic;
    D : in std_logic;
    Q : out std_logic);
end component;

  -- Setting the iob attribute to "true" ensures that instances of these
  -- components are placed inside the I/O pads and are therefore very fast

attribute iob : string;
attribute iob of FD : component is "true";

begin

itrdy_buf : OBUF port map  (
  O => ITRDY,
  I => buf_itrdy
);

itri_buf : OBUF port map  (
  O => ITRI,
  I => buf_itri
);

vbuf : IBUFG port map (
  I => ICLK,
  O => buf_iclk
);

ipgh_pinbuf : IBUF port map (
  I => IPGH,
  O => buf_ipgh
);

ipgh_pinlatch : FD port map (
  C => buf_iclk,
  D => buf_ipgh,
  Q => latched_ipgh
);

ipgv_pinbuf : IBUF port map (
  I => IPGV,
  O => buf_ipgv
);

ipgv_pinlatch : FD port map (
  C => buf_iclk,
  D => buf_ipgv,
  Q => latched_ipgv
);

idq_pinbuf : IBUF port map (
  I => IDQ,
  O => buf_idq
);

idq_pinlatch : FD port map (
  C => buf_iclk,
  D => buf_idq,
  Q => latched_idq
);

databus : for i in 0 to 7 generate
  I_data_pad : IBUF port map (
      I => IPD(i),
      O => buf_ipd(i));

  I_data_ff : FD port map (
      C => buf_iclk,
      D => buf_ipd(i),
      Q => latched_ipd(i));

  H_data_pad : IBUF port map (
      I => HPD(i),
      O => buf_hpd(i));

  H_data_ff : FD port map (
      C => buf_iclk,
      D => buf_hpd(i),
      Q => latched_hpd(i));
end generate;

databus_out : for j in 0 to 15 generate
  regis_count1 : regis16 port map (
    rst    => OPB_Rst,
    clk    => ICLK_OUT,
    load   => load_regs(j+1),
    input  => block_count,
    output => grid_num(j));
end generate;

d1 : demux_1_16
port map (
  sels   => load,
  output => load_regs
);

m1 : mux_16_1
port map (
  i0 => grid_num(0),
  i1 => grid_num(1),
  i2 => grid_num(2),
  i3 => grid_num(3),
  i4 => grid_num(4),
  i5 => grid_num(5),
  i6 => grid_num(6),
  i7 => grid_num(7),
  i8 => grid_num(8),
  i9 => grid_num(9),
  i10 => grid_num(10),
  i11 => grid_num(11),
  i12 => grid_num(12),
  i13 => grid_num(13),
  i14 => grid_num(14),
  i15 => grid_num(15),
  z => data_from_regs,
  sels => load_data
);

u1 : video_decoder_intf
port map (
    ipd           => latched_ipd,
    hpd           => latched_hpd,
    idq           => latched_idq,
    iclk_in       => buf_iclk,
    ipgh          => latched_ipgh,
    ipgv          => latched_ipgv,
    data          => data_from_decoder,
    waddr         => waddr,
    idq_out       => intf_idq_out,
    iclk_out      => intf_iclk_out,
    fil_level     => filling_level,
    line_count    => line_count,
    frame_id      => frame_counter,
    pixel_counter => pix_count,
    active_out    => active,
    reset         => OPB_Rst
);

u2 : video_processing
  port map (
    IPD             => latched_ipd,
    HPD             => latched_hpd,
    IDQ             => latched_idq,
    ICLK            => buf_iclk,
    ICLK_OUT        => ICLK_OUT,
    pix_count       => pix_count,
    line_count      => line_count,
    block_count     => block_count,
    load            => load,
    reset           => OPB_Rst,
    threshold       => threshold
    );

-- Chip select for block RAM - we use existing construct
cs <= OPB_select when OPB_ABus(31 downto 14) = "000000011000000000" else '0';


-- Latching read address. Used to address port A of block RAMs
process (OPB_Clk)
begin
  if OPB_Clk'event and OPB_Clk = '1' then
    if OPB_Rst = '1' then
      raddr <= "000000000000";
    else
      raddr <= OPB_ABus(13 downto 2);
    end if;
  end if;
end process;


-- Latching RNW signal
process (OPB_Clk)
begin
  if OPB_Clk'event and OPB_Clk = '1' then
    if OPB_Rst = '1' then
      rnw <= '0';
    else
      rnw <= OPB_RNW;
    end if;
  end if;
end process;

process(OPB_Clk)
  begin
     if OPB_Clk'event and OPB_Clk='1' then
       if q2='1' and q1='0' and rnw='0' and raddr(11)='1' and raddr(10)='0' then
         threshold <= OPB_DBus;
       end if;
     end if;
  end process;



-- State machine designe for block Ram: we use existing construct
process (OPB_Clk)
begin
  if OPB_Clk'event and OPB_Clk='1' then
    q2 <= (not q2 and q1) or (q2 and not q1);
    q1 <= (cs and not q2 and not q1) or (q2 and not q1);
    q0 <= q2 and not q1;
  end if;
end process;


-- CE is data latch enable
ce <= q2 and not q1 and rnw;


process (OPB_Clk, OPB_Rst)
begin
  if OPB_Rst='1' then
    data_bus_ce <= X"00000000";
  elsif OPB_Clk'event and OPB_Clk='1' then
    if ce='1' then
      if raddr(11)='1' and raddr(10)='1' then
        data_bus_ce <= X"0000000" & filling_level;
      elsif raddr(11)='1' and raddr(10)='0' then
        data_bus_ce <= "0000000000000000000000000000000" & latched_ipgv;
      elsif raddr(11)='0' and raddr(10)='1' then
        data_bus_ce <= X"0000" & line_count;
      elsif raddr(11)='0' and raddr(10)='0' and raddr(9)='1' then
        data_bus_ce <= "0000000000000000000000000000000" & frame_counter(0);
      else
        data_bus_ce <= X"0000" & data_from_regs;
      end if;
    else
      data_bus_ce <= X"00000000";
    end if;
  end if;
end process;


load_data <= raddr(3 downto 0);

-- XFER is transfer acknowledge
xfer <= q0;

-- Slave data bus
VID_DBus <= data_bus_ce;   --VID_DBus(31 downto 0) = VID_DBus

-- Tie unused signals to zero
VID_errAck <= '0';
VID_retry <= '0';
VID_toutSup <= '0';

VID_xferAck <= xfer;

buf_itri <= '1';
buf_itrdy <= '1';

end structural;
